from random import randrange
import numpy as np
from typing import Iterable, Tuple, Dict, Hashable, Sequence

from smqtk_detection.interfaces.detect_image_objects import DetectImageObjects
from smqtk_detection.utils.bbox import AxisAlignedBoundingBox


class RandomDetector(DetectImageObjects):
    """
    Example implementation of the `DetectImageObjects` interface. An instance
    of this class acts as a functor to generate paired bounding boxes and
    classification maps for objects detected in a set of images.
    """

    def detect_objects(
        self,
        img_iter: Iterable[np.ndarray]
    ) -> Iterable[Iterable[Tuple[AxisAlignedBoundingBox, Dict[Hashable, float]]]]:
        """
        Return random set of detections for each image in the input set.
        """

        dets = []

        for img in img_iter:
            img_h = img.shape[0]
            img_w = img.shape[1]

            num_dets = randrange(10)

            dets.append(
                [(
                    self._gen_random_bbox(img_w, img_h),
                    self._gen_random_class_map([0, 1, 2])
                ) for _ in range(num_dets)]
            )

        return dets

    def _gen_random_bbox(self, img_w: int, img_h: int) -> AxisAlignedBoundingBox:
        """
        Creates `AxisAlignedBoundingBox` object with random vertices within
        passed image size.
        """

        min_vertex = [randrange(int(img_w/2)), randrange(int(img_h/2))]
        max_vertex = [randrange(int(img_w/2), img_w), randrange(int(img_h/2), img_h)]

        return AxisAlignedBoundingBox(min_vertex, max_vertex)

    def _gen_random_class_map(self, classes: Sequence) -> Dict[Hashable, float]:
        """
        Creates dictionary of random classification scores for the list of
        input classes.
        """

        scores = np.random.rand(len(classes))
        scores = scores / scores.sum()

        d = {}
        for i, c in enumerate(classes):
            d[c] = scores[i]

        return d

    def get_config(self) -> dict:
        return {}



import abc
from typing import Iterable, Hashable, Dict, Tuple
import numpy as np

from smqtk_core import Configurable, Pluggable
from smqtk_detection.utils.bbox import AxisAlignedBoundingBox


class DetectImageObjects (Configurable, Pluggable):
    """
    Algorithm that generates object bounding boxes and classification maps for
    a set of input image matricies as ``numpy.ndarray`` type arrays.
    """

    @abc.abstractmethod
    def detect_objects(
      self,
      img_iter: Iterable[np.ndarray]
    ) -> Iterable[Iterable[Tuple[AxisAlignedBoundingBox, Dict[Hashable, float]]]]:
        """
        Generate paired bounding boxes and classification maps for detected
        objects in the given set of images.

        :param img_iter: Iterable of input images as numpy arrays.

        :return: Iterable of sets of paired bounding boxes and classification
            maps. Each set is the collection of detections for the
            corresponding input image.
        """

    def __call__(
      self,
      img_iter: Iterable[np.ndarray]
    ) -> Iterable[Iterable[Tuple[AxisAlignedBoundingBox, Dict[Hashable, float]]]]:
        """
        Calls `detect_objects() with the given iterable set of images.`
        """

        return self.detect_objects(img_iter)



from smqtk_detection.exceptions import NoDetectionError
from smqtk_detection.interfaces.detection_element import DetectionElement
from smqtk_classifier.interfaces.classification_element import ClassificationElement
from smqtk_detection.utils.bbox import AxisAlignedBoundingBox
from typing import Hashable, Optional, Tuple


class MemoryDetectionElement (DetectionElement):  # lgtm[py/missing-equals]
    """
    In-memory backend of the DetectionElement representation interface.  This
    implementation has no persistence.

    See ``DetectionElement`` for documentation on abstract method implemented
    here.
    """

    __slots__ = ('_bbox', '_classification')

    @classmethod
    def is_usable(cls) -> bool:
        # In-memory implementation does not require any additional
        # dependencies.
        return True

    def __init__(self, uuid: Hashable) -> None:
        super(MemoryDetectionElement, self).__init__(uuid)
        #: :type: None | AxisAlignedMemoryDetectionElementBoundingBox
        self._bbox: Optional[AxisAlignedBoundingBox] = None
        #: :type: None | ClassificationElement
        self._classification: Optional[ClassificationElement] = None

    def __getstate__(self) -> dict:
        return {
            'parent': super(MemoryDetectionElement, self).__getstate__(),
            'bbox': self._bbox,
            'classification': self._classification,
        }

    def __setstate__(self, state: dict) -> None:
        super(MemoryDetectionElement, self).__setstate__(state['parent'])
        self._bbox = state['bbox']
        self._classification = state['classification']

    def get_config(self) -> dict:
        # No additional constructor parameters for in-memory implementation.
        return {}

    def has_detection(self) -> bool:
        # We are a valid detection if our components are non-null and
        # True-evaluation (meaning they have valid contents).
        if self._bbox is None or self._classification is None:
            return False
        else:
            return self._classification.has_classifications()

    def get_bbox(self) -> Optional[AxisAlignedBoundingBox]:
        if not self._bbox:
            raise NoDetectionError("Missing detection bounding box for "
                                   "in-memory detection with UUID {}"
                                   .format(self.uuid))
        return self._bbox

    def get_classification(self) -> Optional[ClassificationElement]:
        if not (self._classification and
                self._classification.has_classifications()):
            raise NoDetectionError("Missing or empty classification for "
                                   "in-memory detection with UUID {}"
                                   .format(self.uuid))
        return self._classification

    def get_detection(self) -> Tuple[AxisAlignedBoundingBox, ClassificationElement]:
        if (
            self._bbox is None
            or self._classification is None
            or not self._classification.has_classifications()
        ):
            raise NoDetectionError("Missing detection bounding box or "
                                   "missing/invalid classification for "
                                   "in-memory detection with UUID {}"
                                   .format(self.uuid))
        return self._bbox, self._classification

    def set_detection(self, bbox: Optional[AxisAlignedBoundingBox],
                      classification_element: Optional[ClassificationElement]) -> \
            "MemoryDetectionElement":
        if not isinstance(bbox, AxisAlignedBoundingBox):
            raise ValueError("Provided an invalid AxisAlignedBoundingBox "
                             "instance. Given '{}' (type={})."
                             .format(bbox, type(bbox).__name__))
        if not isinstance(classification_element, ClassificationElement):
            raise ValueError("Provided an invalid ClassificationElement "
                             "instance. Given '{}' (type={})."
                             .format(classification_element,
                                     type(classification_element).__name__))
        if not classification_element.has_classifications():
            raise ValueError("Given an empty ClassificationElement instance.")
        self._bbox = bbox
        self._classification = classification_element
        return self



from smqtk_core import Configurable

from smqtk_detection.interfaces.detection_element import DetectionElement
from smqtk_core.configuration import (
    cls_conf_from_config_dict,
    cls_conf_to_config_dict,
    make_default_config,
)
from smqtk_core.dict import merge_dict
from typing import Type, Hashable


class DetectionElementFactory (Configurable):
    """
    Factory class for producing DetectionElement instances of a specified type
    and configuration.
    """

    __slots__ = ('_elem_type', '_elem_config')

    @classmethod
    def get_default_config(cls) -> dict:
        # Override from Configurable
        return make_default_config(DetectionElement.get_impls())

    @classmethod
    def from_config(cls, config_dict: dict, merge_default: bool = True) -> "DetectionElementFactory":
        # Override from Configurable
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        elem_type, elem_conf = cls_conf_from_config_dict(
            config_dict, DetectionElement.get_impls()
        )
        return DetectionElementFactory(elem_type, elem_conf)

    def __init__(self, elem_type: Type[DetectionElement], elem_config: dict) -> None:
        """
        Initialize the factory to produce DetectionElement instance of the
        given type and configuration.

        :param type[DetectionElement] elem_type:
            Instantiable type of DetectionElement this factory should produce.
        :param dict elem_config:
            JSON-compliant dictionary configuration that is to be supplied to
            the input type's ``from_config`` class method. If the ``uuid`` key
            happens to be present in this dictionary it will be ignored
        """
        self._elem_type = elem_type
        self._elem_config = elem_config

    def get_config(self) -> dict:
        return cls_conf_to_config_dict(self._elem_type, self._elem_config)

    def new_detection(self, uuid: Hashable) -> DetectionElement:
        """
        Create a new DetectionElement instance o the configured implementation.

        :param collections.abc.Hashable uuid:
            UUID to assign the element.

        :return: New DetectionElement instance.
        :rtype: DetectionElement

        """
        # noinspection PyUnresolvedReferences
        return self._elem_type.from_config(self._elem_config, uuid)

    __call__ = new_detection
